#!/usr/bin/perl
#
# green_scaled_averaged_harvest_with_QC.pl
#
# This program traverse directories holding GenePix Report files (or
# similar files if you modify it) to harvest, normalize and condense
# the calibration data into one file for subsequent data analysis.
#
# COPYRIGHT AND LICENSE
#
# Copyright 2009 (c) Hui-Hsien Chou
#
# This code is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation. This code is distributed in
# the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License v3 at
# <http://www.gnu.org/licenses/licenses.html> for more details.
#
# To obtain the latest version visit <http://www.complex.iastate.edu>

use strict;
use warnings;

# this module is used to traverse your directory structure
use File::Find;

# this table matches the GPR file column names and orderly index them;
# if yours are different, you need to modify this table.
my %fields = (
"Block" => 0,
"Column" => 1,
"Row" => 2,
"Name" => 3,
"ID" => 4,
"X" => 5,
"Y" => 6,
"Dia." => 7,
"F635 Median" => 8,
"F635 Mean" => 9,
"F635 SD" => 10,
"F635 CV" => 11,
"B635" => 12,
"B635 Median" => 13,
"B635 Mean" => 14,
"B635 SD" => 15,
"B635 CV" => 16,
"% > B635+1SD" => 17,
"% > B635+2SD" => 18,
"F635 % Sat." => 19,
"F532 Median" => 20,
"F532 Mean" => 21,
"F532 SD" => 22,
"F532 CV" => 23,
"B532" => 24,
"B532 Median" => 25,
"B532 Mean" => 26,
"B532 SD" => 27,
"B532 CV" => 28,
"% > B532+1SD" => 29,
"% > B532+2SD" => 30,
"F532 % Sat." => 31,
"Ratio of Medians (635/532)" => 32,
"Ratio of Means (635/532)" => 33,
"Median of Ratios (635/532)" => 34,
"Mean of Ratios (635/532)" => 35,
"Ratios SD (635/532)" => 36,
"Rgn Ratio (635/532)" => 37,
"Rgn R2 (635/532)" => 38,
"F Pixels" => 39,
"B Pixels" => 40,
"Circularity" => 41,
"Sum of Medians (635/532)" => 42,
"Sum of Means (635/532)" => 43,
"Log Ratio (635/532)" => 44,
"F635 Median - B635" => 45,
"F532 Median - B532" => 46,
"F635 Mean - B635" => 47,
"F532 Mean - B532" => 48,
"F635 Total Intensity" => 49,
"F532 Total Intensity" => 50,
"SNR 635" => 51,
"SNR 532" => 52,
"Flags" => 53,
"Normalize" => 54,
"Autoflag" => 55,
"Locus" => 56,
"Annotation" => 57,
);

# interested gene targets; this table translates probe names into
# set/probe encoding pairs; again, you may need to modify this table
# to work on your own data. Note the last line indicates that there
# are two names for the same hygromycin control; we used both.
my %targets = (
"TR035896" => "s1t1", "TR032559" => "s1t2", "TR034678" => "s1t3",
"TR044438" => "s1t4", "TR065861" => "s1t5", "TR073018" => "s1t6",
"TR071427" => "s1t7", "TR041051" => "s1t8", "TR044402" => "s1t9",
"TR044829" => "s1t10",

"TR073011" => "s2t1", "TR060067" => "s2t2", "TR060657" => "s2t3",
"TR047877" => "s2t4", "TR059403" => "s2t5", "TR042189" => "s2t6",
"TR067991" => "s2t7", "TR070396" => "s2t8", "TR041519" => "s2t9",
"TR057842" => "s2t10",

"TR069235" => "s2n1", "TR061421" => "s2n2", "TR035187" => "s2n3",
"TR051070" => "s2n4", "TR062470" => "s2n5", "TR042187" => "s2n6",
"TR054921" => "s2n7", "TR033747" => "s2n8", "TR053363" => "s2n9",
"TR033147" => "s2n10",

"AF354045.1:c9973-8948" => "hygro_control", "TR070096" => "hygro_self"
);

# control gene list; this is similar to the above table but limited to
# only the control probe names; again, modify the table if necessary.
my %c_list = ( "AF354045.1:c9973-8948" => "hygro_control", "TR070096" => "hygro_self");

# feature variable declaration and reference; this table provided a
# declaration of Perl variables that will hold extracted field values
# per each probe that we are interested in the subsequent analysis
my @v_list = \my (
  $name,
  $f635_c,
  $f532_c,
  $ratio, $rgn_ratio, $rgn_r2,
  $b635_2sd, $b532_2sd, $snr_635, $snr_532,
  $flag
);

# feature actually used by the analysis; this matches the table above
# and mention the GPR file columns that we are going to extract into
# the above variables for each probe
my @f_list = @fields{
  "Name",
  "F635 Median - B635",
  "F532 Median - B532",
  "Ratio of Medians (635/532)", "Rgn Ratio (635/532)", "Rgn R2 (635/532)",
  "% > B635+2SD", "% > B532+2SD", "SNR 635", "SNR 532",
  "Flags"
};

# maps feature names to feature variables; this Perl function is a
# "hack" to neatly copy the data fields into declared variables.
sub copy_to_variables {
  foreach (@v_list) {
    $$_ = shift;
  }
}

# need the following on Non-Windows machine ONLY
# local $/ = "\r\n"; # DOS line end is two characters

# the following is the actual subroutine that acts on each GPR file to
# extract its data. Note importantly the pattern matching rule in the
# elsif line below; it extracts temperature, experiment, repeat,
# chip_id, Cy5 and Cy3 PMT settings directly from the file name. If
# your files are organized or named differently, you should modify
# this pattern rule to match your specific data. Note also if an 'x'
# is placed on a file or directory name, it will be skipped. You may
# use this to skip your bad data without removing them. you may also
# want to read the File::Find Perl module document to understand how
# it works and the content of the $File::Find::name variable. In the
# output, all hygromycin intensity values are averaged per each GPR
# file. Similarly, the snr and 2sd are their median values.
my $files = 0;
sub process_files {
  if (/^x /) {
    $File::Find::prune = 1;
  } elsif ($File::Find::name =~ m:(\d+)C_E(\d)_R(\d)_Chip(\d+)_Cy5_(\d+)_Cy3_(\d+)\.gpr:) {
    print STDERR "$File::Find::name\n";
    $files++;
    my $file_name = $_;
    my $degree = $1;
    my $exprim = $2;
    my $repeat = $3;
    my $chip_id = $4;
    my $Cy5_pmt = $5;
    my $Cy3_pmt = $6;

    my @table;
    my ($start, $count, $scale, $red, $green) = ( 0, 0, 0.0, 0.0, 0.0 );
    my @b635_sds;
    my @b532_sds;
    my @snr_635s;
    my @snr_532s;
    open DATA, $file_name;
    while (<DATA>) {
      if ($start) {
	chomp;
	s/\"//g;
	my @f = split /\t/;
	&copy_to_variables(@f[@f_list]);
	next unless $flag>=0 && exists $targets{$name};
	if (exists $c_list{$name}) {
	  if ($rgn_ratio>=0.1 && $rgn_ratio<=10 && $b635_2sd>70 && $b532_2sd>70 &&
	      $ratio>0.8*$rgn_ratio && $ratio<1.2*$rgn_ratio && $rgn_r2>0.5) {
	    $count++;
	    $scale += log($rgn_ratio);
	    $red += $f635_c;
	    $green += $f532_c;
	    push @b635_sds, $b635_2sd;
	    push @b532_sds, $b532_2sd;
	    push @snr_635s, $snr_635;
	    push @snr_532s, $snr_532;
	  }
	} else {
	  push @table, [ $targets{$name}, $f635_c, $f532_c, $b635_2sd, $b532_2sd, $snr_635, $snr_532 ];
	}
      } elsif (/^"Block"/) {
	$start = 1;
      }
    }
    close DATA;
    $scale = exp($scale/$count);
    $red /= $count*$scale;
    $green /= $count;
    my $median = $count/2;
    @b635_sds = sort @b635_sds;
    @b532_sds = sort @b532_sds;
    @snr_635s = sort @snr_635s;
    @snr_532s = sort @snr_532s;
    print join("\t",
	       ($degree, $exprim, $chip_id, "hygro", $Cy5_pmt, $red, $Cy3_pmt, $green,
		$b635_sds[$median], $b532_sds[$median], $snr_635s[$median], $snr_532s[$median])), "\n";
    print STDERR "$count controls, ", scalar @table, " targets, ", $red/$green, " control ratio\n";
    foreach my $t (@table) {
      if ($t->[1]>0) {
	$t->[1] /= $scale;
      } else {
	$t->[1] = 0;
      }
      unless ($t->[2]>=0) {
	$t->[2] = 0;
      }
      print join("\t", ($degree, $exprim, $chip_id, $t->[0], $Cy5_pmt, $t->[1], $Cy3_pmt,
			@{$t}[2..6])), "\n";
    }
  }
}

# this prints the first header line to name the data columns
print join("\t", ('Degree', 'Expr', 'ChipID', 'Probe', 'Cy5_PMT', 'Red',
		  'Cy3_PMT', 'Green', 'Cy5_2SD', 'Cy3_2SD', 'Cy5_SNR', 'Cy3_SNR')), "\n";

# this find command starts the whole traversal; of course you need to
# point the directory argument to the root directory of your data files.
find(\&process_files, "d:/calibration_data");
print STDERR "$files files processed.\n";
