#!/usr/bin/perl
#
# sel1.pl
#
# This program takes a modified Picky output file with added probe
# names and the original gene set given to Picky to design the
# microarray. It then follows the algorithm described in the paper to
# select set 1 calibration probes that satisfy certain conditions. The
# antisense oligo pair for each probe are also selected. See the paper
# for more details about its algorithm.
#
# Note that this Perl program requires the Align module supplied.
#
# COPYRIGHT AND LICENSE
#
# Copyright 2009 (c) Hui-Hsien Chou
#
# This code is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation. This code is distributed in
# the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License v3 at
# <http://www.gnu.org/licenses/licenses.html> for more details.
#
# To obtain the latest version visit <http://www.complex.iastate.edu>

use strict;
use warnings;

use Align;

# step 1: record all target names from TIGR choices
my $i;
my $d;
my %table;  # complete table of chosen oligos
my %list;   # partial list of usable oligos
open TIGR, "$ARGV[0]";
for ($i=$d=0; <TIGR>; $i++) {
  if ($i) {
    chomp;
    my @fields = split/\t/;
    for (my $j=8; $j<@fields; $j+=4) {
      unless (exists $table{$fields[$j]}) {
	$table{$fields[$j]} = 
	    [$fields[$j], $fields[0], @fields[3, $j+2, $j+3],
	     $fields[4], @fields[7, 6, 1]];
	$list{$fields[$j]} = $table{$fields[$j]} if $fields[5] =~ />/;
	$d++;
      }
    }
  }
}
close TIGR;
print STDERR "Step 1: --------------------\n";
print STDERR "TIGR chosen oligos = ", $i-1, "\n";
print STDERR "Target sequences = $d\n";

# step 2: keep only target/nontarget list that are both represented on the array
$d = 0;
foreach $i (values %list) {
  unless (exists $table{${$i}[5]}) {
    delete $list{${$i}[0]};
  } else {
    $d++;
  }
}
print STDERR "Step 2: --------------------\n";
die "mismatch hash table count" if $d != scalar keys %list;
print STDERR "Hash table entry count = $d\n";

# step 3: record gene sequence locations
my $n;
my %map;
open GENES, "$ARGV[1]";
for ($i=0; <GENES>; ) {
  if (/^>/) {
    my @fields = split;
    $n = substr($fields[0], 1);
    $map{$n} = tell GENES if exists $table{$n};
    $i++;
  }
}
print STDERR "Step 3: --------------------\n";
print STDERR "Rice total gene count = $i\n";

# step 4: sort target/nontarget list based on temperature difference
# list array content:
# [target, t_id, t_temp, t_left, t_right, nontarget, n_temp, n_site, oligo]
# [     0,    1,      2,      3,       4,         5,      6,      7,     8]
sub by_temp_diff {
  my $x = ${$a}[2]-${$a}[6];
  my $y = ${$b}[2]-${$b}[6];
  if ($x<$y) { -1 } elsif ($x>$y) { 1 } else { 0 }
}
my @target = sort by_temp_diff values %list;

# redirect %list to map to sorted @target indices
for ($i=0; $i<@target; $i++) {
  $list{${$target[$i]}[0]} = $i;
}
print STDERR "Step 4: --------------------\n";
print STDERR "Target/nontarget pair count = $i\n";

sub rev_comp {
  my $r = scalar reverse $_[0];
  $r =~ tr/acgtACGT/TGCATGCA/;
  return $r;
}

# step 5: produce output for melting temperature calibration
my $count=0;
print "oligo_id\toligo\to_left\to_right\ttarget\tt_temp\tt_left\tt_right\tt_anti_oligo";
print "\tnontarget\tnt_oligo_id\tn_temp\tn_left\tn_right\tn_anti_oligo\n";
for ($i=0; $i<@target && $count<1000; $i++) {
  my $x = $target[$i];
  next if exists($list{${$x}[5]}) && $list{${$x}[5]}<=$i;

  # output: oligo_id, oligo, o_left, o_right, target, t_temp
  my $output = join "\t", @{$x}[1, 8, 3, 4, 0, 2];

  # load target gene sequence from input sequence file
  seek GENES, $map{${$x}[0]}, 0;
  for ($d = ""; <GENES>; ) {
    last unless $_ !~ /^>/;
    chomp;
    $d .= $_;
  }

  # align target probe with target and double check coordinate consistency
  my ($t_left, $t_right, $p_left, $p_right);
  Align::Align($d, ${$x}[8], $t_left, $t_right, $p_left, $p_right);
  die "mismatch left target coordinate $t_left!=${$x}[3]" if $t_left!=${$x}[3]; 
  die "mismatch right target coordinate $t_right!=${$x}[4]" if $t_right!=${$x}[4];

  # adjust anti-sense location for possible inclusion of dangling ends
  # extract anti-sense fragment from the target sequence
  $t_left-- if $t_left;
  $t_right++;
  $t_right++ if $t_right<length($d);
  my $s = &rev_comp(substr($d, $t_left, $t_right-$t_left));

  # output: t_left, t_right, t_anti_oligo
  $output .= "\t$t_left\t".($t_right-1)."\t$s";

  # load nontarget gene sequence from input file
  my $y = $table{${$x}[5]};
  die "mismatch nontarget names ${$y}[0]!=${$x}[5]" if ${$y}[0] ne ${$x}[5];
  seek GENES, $map{${$y}[0]}, 0;
  for ($n = ""; <GENES>; ) {
    last unless $_ !~ /^>/;
    chomp;
    $n .= $_;
  }

  # output: nontarget, nt_oligo_id, n_temp
  $output .= "\t${$y}[0]\t${$y}[1]\t${$x}[6]";

  # align nontarget probe with nontarget and double check coordinate consistency
  Align::Align($n, ${$y}[8], $t_left, $t_right, $p_left, $p_right);
  die "mismatch left nontarget coordinate $t_left!=${$y}[3]" if $t_left!=${$y}[3]; 
  die "mismatch right nontarget coordinate $t_right!=${$y}[4]" if $t_right!=${$y}[4];

  # align target probe with the nontarget sequence
  Align::Align($n, ${$x}[8], $t_left, $t_right, $p_left, $p_right);

  # we prefer only nontarget match centered in the middle of the oligo
  next unless abs($p_left+$p_right-length(${$x}[8]))<=30;

  # adjust anti-sense location for possible inclusion of dangling ends
  $t_left -= $p_left+1;
  $t_left = 0 if $t_left<0;
  $t_right = $t_left + ${$x}[4]-${$x}[3] + 3;
  if ($t_right > length($n)) {
    $t_right = length($n);
    $t_left = $t_right - ${$x}[4]+${$x}[3] - 3;
  }

  # for set 1, only choose nontargets with non-overlapping alignment site
  if ($t_left>${$y}[4] || $t_right<${$y}[3]) {

    # extract anti-sense fragment from the nontarget sequence
    $s = &rev_comp(substr($n, $t_left, $t_right-$t_left));

    # output: n_left, n_right, n_anti_oligo
    $output .= "\t$t_left\t".($t_right-1)."\t$s";

    # actually producing the output
    print $output, "\n";
    $count++;
  }
}
print STDERR "Step 5: --------------------\n";
print STDERR "$i target/nontarget pairs examined, $count pairs selected\n";
print STDERR scalar @target, " total candidates.\n";
close GENES;
